@@ -136,4 +136,10 @@ try:
 except:
     # for orchestrationmanager to be able to come up independently
     # TODO: fix this
-    pass
\ No newline at end of file
+    pass
+
+
+class FatalError(Exception):
+    """
+    raised to exit pytest immediately
+    """
\ No newline at end of file
