@@ -1,15 +1,13 @@
 from dataclasses import dataclass
 from enum import Enum
-from typing import List, Any, Set, Dict
+from typing import List, Union, Optional
 
-from _pytest._code.code import ReprExceptionInfo
-from _pytest.reports import TestReport
-
-
-DEFAULT_PRIORITY = 10
+from pytest_mproc import DEFAULT_PRIORITY
+from pytest import TestReport
 
 
 class TestExecutionConstraint(Enum):
+    __test__ = False  # tell pytest not to treat as test class
     SINGLE_PROCESS ="SINGLE_PROCESS"
     SINGLE_NODE = "SINGLE_NODE"
 
@@ -30,8 +28,21 @@ class ResourceUtilization:
 
 
 @dataclass
-class ResultException:
-    excrepr: ReprExceptionInfo
+class ResultException(Exception):
+    excrepr: Exception
+
+
+@dataclass
+class ClientDied(Exception):
+    pid: int
+    host: str
+    errored: bool = False
+    message: Optional[str] = None
+
+
+@dataclass
+class AllClientsCompleted:
+    pass
 
 
 @dataclass
@@ -39,6 +50,20 @@ class ResultTestStatus:
     report: TestReport
 
 
+class TestStateEnum(Enum):
+    STARTED = 1
+    FINISHED = 2
+    RETRY = 3
+
+
+@dataclass
+class TestState:
+    state: TestStateEnum
+    host: str
+    pid: int
+    test_id: str
+    test_batch: TestBatch
+
 @dataclass
 class ResultExit:
     worker_index: int
@@ -56,3 +81,6 @@ class GroupTag:
 
     def __hash__(self):
         return self.name.__hash__()
+
+
+ResultType = Union[TestState, ResultException, ResultExit, ResourceUtilization, ResultTestStatus]
